"""Various utilities for building the HTML log file."""

# © 2023 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS).  Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

# SPDX-License-Identifier: BSD-3-Clause

import pkgutil
import re
import textwrap
from collections.abc import Iterable, Mapping
from datetime import datetime
from pathlib import Path
from types import SimpleNamespace
from typing import Iterator, TextIO, Union


def nested_simplenamespace_to_dict(
    namespace: Union[str, bytes, tuple, Mapping, Iterable, SimpleNamespace],
) -> Union[str, bytes, tuple, dict, list]:
    """
    Convert a ``SimpleNamespace`` to a ``dict``.

    Convert a ``SimpleNamespace``, which may include nested namespaces,
    iterables, and mappings, to a ``dict`` containing the equivalent
    items.

    Parameters:
        namespace:  The given namespace to convert, or some nested
            element therein.

    Returns:
        Recursively returns a base Python type equivalent of whatever
        was given.
    """
    if "_asdict" in dir(namespace):
        # noinspection PyProtectedMember
        return nested_simplenamespace_to_dict(namespace._asdict())
    if isinstance(namespace, (str, bytes, tuple)):
        return namespace
    if isinstance(namespace, Mapping):
        return {
            k: nested_simplenamespace_to_dict(v) for k, v in namespace.items()
        }
    if isinstance(namespace, Iterable):
        return [nested_simplenamespace_to_dict(x) for x in namespace]
    if isinstance(namespace, SimpleNamespace):
        return nested_simplenamespace_to_dict(namespace.__dict__)
    return namespace


def get_human_time(milliseconds: float) -> str:
    """
    Get a human-readable date/time.

    Parameters:
        milliseconds:  The number of milliseconds since epoch.

    Returns:
        A string representation of the date and time.
    """
    seconds = milliseconds / 1000.0
    return datetime.fromtimestamp(seconds).strftime("%Y-%m-%d %H:%M:%S.%f")


def opening_html_text() -> str:
    """
    Get the opening HTML text.

    Returns:
        A string containing the first line of the HTML document through
        ``</head>``.
    """
    return "<!DOCTYPE html><html>" + html_header()


def closing_html_text() -> str:
    """
    Get the closing HTML tag.

    Returns:
        A string with the closing HTML tag in it.
    """
    return "</html>"


def append_html(*args: Union[str, Iterator[str]], output: Path) -> None:
    """
    Append whatever is given to the ``output`` HTML file.

    Parameters:
        *args:  The argument(s) to write.
        output:  The HTML file to append to.
    """

    def _append_html(
        f: TextIO, *inner_args: Union[str, bytes, Iterable]
    ) -> None:
        """
        Write some text to the given HTML log file.

        Parameters:
            f:  The HTML file to write to.
            *inner_args:  The argument(s) to write.
        """
        for arg in inner_args:
            if isinstance(arg, str):
                f.write(arg)
            elif isinstance(arg, bytes):
                f.write(arg.decode())
            elif isinstance(arg, Iterable):
                _append_html(f, *arg)
            else:
                message = f"Unsupported type: {type(arg)}"
                raise TypeError(message)

    with output.open("a") as output_file:
        _append_html(output_file, *args)


def fixed_width(text: str) -> str:
    """
    Convert the text to a fixed-width font.

    Wrap the given ``text`` in a ``<pre><code>...</code></pre>`` block
    such that it displays in a fixed-width font.

    Parameters:
        text:  The text to wrap.

    Returns:
        The ``<pre><code>...</code></pre>`` block.
    """
    return f"<pre><code>{html_encode(text)}</code></pre>"


def flatten(element: Union[str, bytes, Iterable]) -> Iterator[str]:
    """
    Turn a tree of lists into a flat iterable of strings.

    Parameters:
        element:  An element of a tree.

    Yields:
        The string representation of the given element.
    """
    if isinstance(element, str):
        yield element
    elif isinstance(element, bytes):
        yield element.decode()
    elif isinstance(element, Iterable):
        for _element in element:
            yield from flatten(_element)
    else:
        yield element


def parent_logger_card_html(
    name: str, *args: list[Iterator[str]]
) -> Iterator[str]:
    """
    Generate the HTML for a parent logger card.

    Generate the HTML for the card corresponding to the parent
    :class:`ShellLogger`.  The HTML elements are yielded one at a time
    to avoid loading *all* the data from the :class:`ShellLogger` into
    memory at once.

    Parameters:
        name:  The name of the :class:`ShellLogger`.
        *args:  A list of generators to lazily yield string HTML
            elements for the contents of the parent card.

    Yields:
        The header, followed by all the contents of the
        :class:`ShellLogger`, and then the footer.
    """
    header, indent, footer = split_template(
        parent_logger_template, "parent_body", name=name
    )
    yield header
    for arg in flatten(args):
        yield textwrap.indent(arg, indent)
    yield footer


def child_logger_card(log) -> Iterator[str]:
    """
    Generate a child logger card.

    Create a card to go in the HTML log file containing everything
    pertaining to a child :class:`ShellLogger`.

    Parameters:
        log (ShellLogger):  The child :class:`ShellLogger` for which to
            generate the card.

    Returns:
        A generator that will lazily yield the elements of the HTML for
        the card one at a time.

    Todo:
        * The type hinting for ``log`` is done in the docstring instead
          of the function signature, because to put it in the signature
          would create a circular dependency between ``shell_logger.py``
          and ``html_utilities.py``.  This function needs to be reworked
          such that there's no longer a dependency on ``ShellLogger``.
    """
    child_html = log.to_html()
    return child_logger_card_html(log.name, log.duration, *child_html)


def child_logger_card_html(
    name: str, duration: str, *args: Union[Iterator[str], list[Iterator[str]]]
) -> Iterator[str]:
    """
    Generate the HTML for a child logger card.

    Generate the HTML for a card corresponding to the child
    :class:`ShellLogger`.  The HTML elements are yielded one at a time
    to avoid loading *all* the data from the :class:`ShellLogger` into
    memory at once.

    Parameters:
        name:  The name of the child :class:`ShellLogger`.
        duration:  The duration of the child :class:`ShellLogger`.
        *args:  A generator (or list of generators) to lazily yield
            string HTML elements for the contents of the child card.

    Yields:
        The header, followed by all the contents of the child
        :class:`ShellLogger`, and then the footer.

    Todo:
      * Should we replace the ``for`` loop with the one found in
        :func:`parent_logger_card_html`?
    """
    header, indent, footer = split_template(
        child_logger_template, "child_body", name=name, duration=duration
    )
    yield header
    for arg in args:
        if isinstance(arg, str):
            yield textwrap.indent(arg, indent)
        elif isinstance(arg, Iterable):
            for _arg in arg:
                yield textwrap.indent(_arg, indent)
    yield footer


def command_card_html(
    log: dict, *args: Iterator[Union[str, Iterable]]
) -> Iterator[str]:
    """
    Generate the HTML for a command card.

    Generate the HTML for a card corresponding to a command that was
    run.  The HTML elements are yielded one at a time to avoid loading
    *all* the data into memory at once.

    Parameters:
        log:  An entry from the :class:`ShellLogger` 's log book
            corresponding to a command that was run.
        *args:  A generator that will yield all the elements to be
            included in the command card one at a time.

    Yields:
        The header, followed by all the contents of the command card,
        and then the footer.
    """
    header, indent, footer = split_template(
        command_template,
        "more_info",
        cmd_id=log["cmd_id"],
        command=fixed_width(log["cmd"]),
        message=log["msg"],
        return_code=log["return_code"],
        duration=log["duration"],
    )
    yield header
    for arg in args:
        if isinstance(arg, str):
            yield textwrap.indent(arg, indent)
        elif isinstance(arg, Iterable):
            for _arg in arg:
                yield textwrap.indent(_arg, indent)
    yield footer


def html_message_card(log: dict) -> Iterator[str]:
    """
    Generate the HTML for a message card.

    Generate the HTML for a card corresponding to a message to only be
    included in the HTML log file (e.g., not printed to ``stdout`` as
    well).

    Parameters:
        log:  An entry from the :class:`ShellLogger` 's log book
            corresponding to a message.

    Yields:
        The header, followed by the contents of the message card, and
        then the footer.
    """
    timestamp = (
        log["timestamp"]
        .replace(" ", "_")
        .replace(":", "-")
        .replace("/", "_")
        .replace(".", "-")
    )
    header, indent, footer = split_template(
        html_message_template,
        "message",
        title=log["msg_title"],
        timestamp=timestamp,
    )
    text = html_encode(log["msg"])
    text = "<pre>" + text.replace("\n", "<br>") + "</pre>"
    yield header
    yield textwrap.indent(text, indent) + "\n"
    yield footer


def message_card(log: dict) -> Iterator[str]:
    """
    Generate a message card.

    Generate the HTML for a card corresponding to a message to be both
    printed to ``stdout`` and included in the HTML log file.

    Parameters:
        log:  An entry from the :class:`ShellLogger` 's log book
            corresponding to a message.

    Yields:
        The header, followed by the contents of the message card, and
        then the footer.
    """
    header, indent, footer = split_template(message_template, "message")
    text = html_encode(log["msg"])
    text = "<pre>" + text.replace("\n", "<br>") + "</pre>"
    yield header
    yield textwrap.indent(text, indent) + "\n"
    yield footer


def command_detail_list(cmd_id: str, *args: Iterator[str]) -> Iterator[str]:
    """
    Generate the list of command details.

    Generate the HTML for a list of details associated with a command
    that was run.

    Parameters:
        cmd_id:  The unique identifier associated with the command that
            was run.
        *args:  All of the details associated with a command that was
            run.

    Yields:
        The header, followed by each of the details associated with the
        command that was run, and then the footer.
    """
    header, indent, footer = split_template(
        command_detail_list_template, "details", cmd_id=cmd_id
    )
    yield header
    for arg in args:
        if isinstance(arg, str):
            yield textwrap.indent(arg, indent)
    yield footer


def command_detail(
    cmd_id: str, name: str, value: str, *, hidden: bool = False
) -> str:
    """
    Generate the HTML for a command detail.

    Create the HTML snippet for a detail associated with a command that
    was run.

    Parameters:
        cmd_id:  The unique identifier associated with the command that
            was run.
        name:  The name of the detail being recorded.
        value:  The value of the detail being recorded.
        hidden:  Whether or not this detail should be hidden (collapsed)
            in the HTML by default.

    Returns:
        The HTML snippet for this command detail.
    """
    if hidden:
        return hidden_command_detail_template.format(
            cmd_id=cmd_id, name=name, value=value
        )
    return command_detail_template.format(name=name, value=value)


def command_card(log: dict, stream_dir: Path) -> Iterator[str]:
    """
    Generate a command card.

    Create a card in the HTML log file containing the output of a
    command, along with all its corresponding data (environment
    information, trace output, memory/CPU/disk statistics, etc.).

    Parameters:
        log:  An entry from the :class:`ShellLogger` 's log book
            corresponding to a command that was run.
        stream_dir:  The stream directory containing the ``stdout``,
            ``stderr``, and ``trace`` output from the command.

    Returns:
        A generator to lazily yield the elements of the command card one
        at a time.
    """
    cmd_id = log["cmd_id"]
    stdout_path = stream_dir / f"{log['timestamp']}_{cmd_id}_stdout"
    stderr_path = stream_dir / f"{log['timestamp']}_{cmd_id}_stderr"
    trace_path = stream_dir / f"{log['timestamp']}_{cmd_id}_trace"

    # Collect all the details associated with the command that was run.
    info = [
        command_detail_list(
            cmd_id,
            command_detail(cmd_id, "Time", log["timestamp"]),
            command_detail(cmd_id, "Command", fixed_width(log["cmd"])),
            command_detail(cmd_id, "CWD", log["pwd"], hidden=True),
            command_detail(cmd_id, "Hostname", log["hostname"], hidden=True),
            command_detail(cmd_id, "User", log["user"], hidden=True),
            command_detail(cmd_id, "Group", log["group"], hidden=True),
            command_detail(cmd_id, "Shell", log["shell"], hidden=True),
            command_detail(cmd_id, "umask", log["umask"], hidden=True),
            command_detail(cmd_id, "Return Code", log["return_code"]),
        ),
        output_block_card("stdout", stdout_path, cmd_id, collapsed=False),
        output_block_card("stderr", stderr_path, cmd_id, collapsed=False),
    ]

    # Compile the additional diagnostic information.
    diagnostics = [
        output_block_card("Environment", log["environment"], cmd_id),
        output_block_card("ulimit", log["ulimit"], cmd_id),
    ]
    if trace_path.exists():
        diagnostics.append(output_block_card("trace", trace_path, cmd_id))

    # Add in any available statistics (from `StatsCollector`s).
    if log.get("stats"):
        stats = [("memory", "Memory Usage"), ("cpu", "CPU Usage")]
        for stat, stat_title in stats:
            if log["stats"].get(stat):
                data = log["stats"][stat]
                diagnostics.append(time_series_plot(cmd_id, data, stat_title))
        if log["stats"].get("disk"):
            uninteresting_disks = [
                "/var",
                "/var/log",
                "/var/log/audit",
                "/boot",
                "/boot/efi",
            ]
            disk_stats = {
                x: y
                for x, y in log["stats"]["disk"].items()
                if x not in uninteresting_disks
            }

            # We sort because JSON deserialization may change
            # the ordering of the map.
            for disk, data in sorted(disk_stats.items()):
                diagnostics.append(disk_time_series_plot(cmd_id, data, disk))
    info.append(diagnostics_card(cmd_id, *diagnostics))
    return command_card_html(log, *info)


def time_series_plot(
    cmd_id: str, data_tuples: list[tuple[float, float]], series_title: str
) -> Iterator[str]:
    """
    Create the HTML for a plot of time series data.

    Parameters:
        cmd_id:  The unique identifier associated with the command that
            was run.
        data_tuples:  A list of :math:`x` and :math:`y` locations.
        series_title:  The title of the plot.

    Returns:
        A HTML snippet for a plot of the given data.
    """
    labels = [get_human_time(x) for x, _ in data_tuples]
    values = [y for _, y in data_tuples]
    identifier = f"{cmd_id}-{series_title.lower().replace(' ', '-')}-chart"
    return stat_chart_card(labels, values, series_title, identifier)


def disk_time_series_plot(
    cmd_id: str, data_tuples: list[tuple[float, float]], volume_name: str
) -> Iterator[str]:
    """
    Generate a time series plot of disk usage.

    Create the HTML for a plot of the disk usage time series data for a
    particular volume.

    Parameters:
        cmd_id:  The unique identifier associated with the command that
            was run.
        data_tuples:  A list of :math:`x` and :math:`y` locations.
        volume_name:  The name of the disk volume who's data is being
            plotted.

    Returns:
        A HTML snippet for a plot of the given data.

    Todo:
      * Should we combine this with :func:`time_series_plot`?
    """
    labels = [get_human_time(x) for x, _ in data_tuples]
    values = [y for _, y in data_tuples]
    identifier = f"{cmd_id}-volume{volume_name.replace('/', '_')}-usage"
    stat_title = f"Used Space on {volume_name}"
    return stat_chart_card(labels, values, stat_title, identifier)


def stat_chart_card(
    labels: list[str], data: list[float], title: str, identifier: str
) -> Iterator[str]:
    """
    Create the HTML for a two-dimensional plot.

    Parameters:
        labels:  The :math:`x` values.
        data:  The :math:`y` values.
        title:  The title for the plot.
        identifier:  A unique identifier for the chart.

    Yields:
        A HTML snippet for the chart with all the details filled in.
    """
    yield stat_chart_template.format(
        labels=labels, data=data, title=title, id=identifier
    )


def output_block_card(
    title: str,
    output: Union[Path, str],
    cmd_id: str,
    *,
    collapsed: bool = True,
) -> Iterator[str]:
    """
    Generate an output block card.

    Given the output from a command, generate a corresponding HTML card
    for inclusion in the log file.

    Parameters:
        title:  The title for the output block.
        output:  The output from a command.
        cmd_id:  The unique identifier associated with the command that
            was run.
        collapsed:  Whether or not the output block should be collapsed
            by default in the HTML log file.

    Yields:
        The header, followed by each line of the output, and then the
        footer.
    """
    name = title.replace(" ", "_").lower()
    template = (
        output_card_collapsed_template if collapsed else output_card_template
    )
    header, indent, footer = split_template(
        template, "output_block", name=name, title=title, cmd_id=cmd_id
    )
    yield header
    for line in output_block(output, name, cmd_id):
        yield textwrap.indent(line, indent)
    yield footer


def output_block(
    output: Union[Path, str], name: str, cmd_id: str
) -> Iterator[str]:
    """
    Generate an output block.

    Given the output from a command, generate the HTML equivalent for
    inclusion in the log file.

    Parameters:
        output:  The output from a command.
        name:  The name (title) of the output block.
        cmd_id:  The unique identifier associated with the command that
            was run.

    Yields:
        The HTML equivalent of each line of the output in turn.
    """
    if isinstance(output, Path):
        with output.open() as f:
            for string in output_block_html(f, name, cmd_id):
                yield string
    if isinstance(output, str):
        for string in output_block_html(output, name, cmd_id):
            yield string


def diagnostics_card(cmd_id: str, *args: Iterator[str]) -> Iterator[str]:
    """
    Generate a diagnostics card.

    Generate a card containing system diagnostic information associated
    with a command that was run.

    Parameters:
        cmd_id:  The unique identifier associated with the command that
            was run.
        *args:  A generator to lazily yield all the diagnostic
            information, one piece at a time.

    Yields:
        The header, followed by each piece of diagnostic information,
        and then the footer.
    """
    header, indent, footer = split_template(
        diagnostics_template, "diagnostics", cmd_id=cmd_id
    )
    yield header
    for arg in args:
        if isinstance(arg, str):
            yield textwrap.indent(arg, indent)
        elif isinstance(arg, Iterable):
            for _arg in arg:
                yield textwrap.indent(_arg, indent)
    yield footer


def output_block_html(
    lines: Union[TextIO, str], name: str, cmd_id: str
) -> Iterator[str]:
    """
    Generate the HTML for an output block.

    Given the output of a command, generate its HTML equivalent for
    inclusion in the log file.

    Parameters:
        lines:  The lines of output.
        name:  The name (title) for this output block.
        cmd_id:  The unique identifier associated with the command that
            was run.

    Yields:
        The header, followed by the HTML corresponding to each line of
        output, and then the footer.
    """
    if isinstance(lines, str):
        lines = lines.split("\n")
    header, indent, footer = split_template(
        output_block_template, "table_contents", name=name, cmd_id=cmd_id
    )
    yield header
    for line_no, line in enumerate(lines):
        yield textwrap.indent(output_line_html(line, line_no), indent)
    yield footer


def split_template(
    template: str, split_at: str, **kwargs
) -> tuple[str, str, str]:
    """
    Subdivide a HTML template.

    Take a templated HTML snippet and split it into a header and footer,
    meaning everything that comes before and after the line containing
    ``split_at``.  Also determine the indentation for the content that
    will be inserted between the header and footer.

    Example:
        If the following snippet is ``template`` and ``split_at`` is
        ``child_body``, then the header is lines 1-6, the footer is
        lines 8-9, and the indent is eight spaces.

        .. code-block:: html
           :linenos:
           :emphasize-lines: 7

           <details class="child-logger">
               <summary>
                   <h6 class="child-logger-heading">{name}</h6>
                   <span class="duration"> (Duration: {duration})</span>
               </summary>
               <div class="child-logger-body">
                   {child_body}
               </div>
           </details>

    Parameters:
        template:  A templated HTML snippet.
        split_at:  A substring used to split the ``template`` into
            before and after chunks.
        **kwargs:  Additional keyword arguments used to replace keywords
            in the ``template``.

    Returns:
        The header, indent, and footer.
    """
    fmt = {k: v for k, v in kwargs.items() if k != split_at}
    pattern = re.compile(
        f"(.*\\n)(\\s*)\\{{{split_at}\\}}\\n(.*)", flags=re.DOTALL
    )
    before, indent, after = pattern.search(template).groups()
    return before.format(**fmt), indent, after.format(**fmt)


def output_line_html(line: str, line_no: int) -> str:
    """
    Generate the HTML for a line of output.

    Given a line of output from a command, along with the corresponding
    line number, create the HTML equivalent to be included in the log
    file.

    Parameters:
        line:  A line of output.
        line_no:  The corresponding line number.

    Returns:
        The corresponding HTML snippet.
    """
    encoded_line = html_encode(line).rstrip()
    return output_line_template.format(line=encoded_line, line_no=line_no)


def html_encode(text: str) -> str:
    """
    Replace special characters with their HTML encodings.

    Parameters:
        text:  The text to encode.

    Returns:
        The encoded text.
    """
    return sgr_to_html(
        text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")
    )


def sgr_to_html(text: str) -> str:
    """
    Convert SGR to HTML.

    Translate Select Graphic Rendition (SGR, a.k.a. ANSI escape codes)
    to valid HTML/CSS.

    Parameters:
        text:  The input text.

    Returns:
        The same text, with the escape codes translated to HTML/CSS.
    """
    span_count = 0
    while text.find("\x1b[") >= 0:
        start = text.find("\x1b[")
        finish = text.find("m", start)
        sgrs = text[start + 2 : finish].split(";")
        span_string = ""
        if len(sgrs) == 0:
            span_string += "</span>" * span_count
            span_count = 0
        else:
            while sgrs:
                if sgrs[0] == "0":
                    span_string += "</span>" * span_count
                    span_count = 0
                    sgrs = sgrs[1:]
                elif len(sgrs) >= 5 and sgrs[:2] in [["38", "2"], ["48", "2"]]:
                    span_count += 1
                    span_string += sgr_24bit_color_to_html(sgrs[:5])
                    sgrs = sgrs[5:]
                elif len(sgrs) >= 3 and sgrs[:2] in [["38", "5"], ["48", "5"]]:
                    span_count += 1
                    span_string += sgr_8bit_color_to_html(sgrs[:3])
                    sgrs = sgrs[3:]
                else:
                    span_count += 1
                    span_string += sgr_4bit_color_and_style_to_html(sgrs[0])
                    sgrs = sgrs[1:]
        text = text[:start] + span_string + text[finish + 1 :]
    return text


def sgr_4bit_color_and_style_to_html(sgr: str) -> str:
    """
    Convert from Select Graphic Rendition (SGR) codes to CSS styles.

    Parameters:
        sgr:  The SGR code specified.

    Returns:
        A HTML ``span`` with the corresponding CSS style definition.
    """
    sgr_to_css = {
        "1": "font-weight: bold;",
        "2": "font-weight: lighter;",
        "3": "font-style: italic;",
        "4": "text-decoration: underline;",
        "9": "text-decoration: line-through;",
        "30": "color: black;",
        "40": "background-color: black;",
        "31": "color: red;",
        "41": "background-color: red;",
        "32": "color: green;",
        "42": "background-color: green;",
        "33": "color: yellow;",
        "43": "background-color: yellow;",
        "34": "color: blue;",
        "44": "background-color: blue;",
        "35": "color: magenta;",
        "45": "background-color: magenta;",
        "36": "color: cyan;",
        "46": "background-color: cyan;",
        "37": "color: white;",
        "47": "background-color: white;",
        "90": "color: black;",
        "100": "background-color: black;",
        "91": "color: red;",
        "101": "background-color: red;",
        "92": "color: green;",
        "102": "background-color: green;",
        "93": "color: yellow;",
        "103": "background-color: yellow;",
        "94": "color: blue;",
        "104": "background-color: blue;",
        "95": "color: magenta;",
        "105": "background-color: magenta;",
        "96": "color: cyan;",
        "106": "background-color: cyan;",
        "97": "color: white;",
        "107": "background-color: white;",
        "39": "color: inherit;",
        "49": "background-color: inherit;",
    }
    return f'<span style="{sgr_to_css.get(sgr) or str()}">'


def sgr_8bit_color_to_html(sgr_params: list[str]) -> str:  # noqa: PLR0911
    """
    Convert 8-bit SGR colors to HTML.

    Convert an 8-bit Select Graphic Rendition (SGR) code to valid
    HTML/CSS.

    Parameters:
        sgr_params:  The SGR codes to convert.

    Returns:
        A HTML ``span`` with the appropriate CSS style.

    Raises:
        RuntimeError:  If the code somehow falls through all the `if`
            blocks without returning.  This should never happen.
    """
    sgr_256 = int(sgr_params[2]) if len(sgr_params) > 2 else 0
    if sgr_256 < 0 or sgr_256 > 255 or not sgr_params:
        return "<span>"
    if 15 < sgr_256 < 232:
        red_6cube = (sgr_256 - 16) // 36
        green_6cube = (sgr_256 - (16 + red_6cube * 36)) // 6
        blue_6cube = (sgr_256 - 16) % 6
        red = str(51 * red_6cube)
        green = str(51 * green_6cube)
        blue = str(51 * blue_6cube)
        return sgr_24bit_color_to_html([sgr_params[0], "2", red, green, blue])
    if 231 < sgr_256 < 256:
        gray = str(8 + (sgr_256 - 232) * 10)
        return sgr_24bit_color_to_html([sgr_params[0], "2", gray, gray, gray])
    if sgr_params[0] == "38":
        if sgr_256 < 8:
            return sgr_4bit_color_and_style_to_html(str(30 + sgr_256))
        if sgr_256 < 16:
            return sgr_4bit_color_and_style_to_html(str(82 + sgr_256))
    if sgr_params[0] == "48":
        if sgr_256 < 8:
            return sgr_4bit_color_and_style_to_html(str(40 + sgr_256))
        if sgr_256 < 16:
            return sgr_4bit_color_and_style_to_html(str(92 + sgr_256))
    message = "THIS SHOULD NEVER HAPPEN"
    raise RuntimeError(message)


def sgr_24bit_color_to_html(sgr_params: list[str]) -> str:
    """
    Convert 24-bit SGR colors to HTML.

    Convert a 24-bit Select Graphic Rendition (SGR) code to valid
    HTML/CSS.

    Parameters:
        sgr_params:  The SGR codes to convert.

    Returns:
        A HTML ``span`` with the appropriate CSS style.
    """
    r, g, b = sgr_params[2:5] if len(sgr_params) == 5 else ("0", "0", "0")
    if len(sgr_params) > 1 and sgr_params[:2] == ["38", "2"]:
        return f'<span style="color: rgb({r}, {g}, {b})">'
    if len(sgr_params) > 1 and sgr_params[:2] == ["48", "2"]:
        return f'<span style="background-color: rgb({r}, {g}, {b})">'
    return "<span>"


def html_header() -> str:
    """
    Get the HTML header, complete with embedded styles and scripts.

    Returns:
        A string with the ``<head>...</head>`` contents.
    """
    return (
        "<head>"
        + embed_style("bootstrap.min.css")
        + embed_style("Chart.min.css")
        + embed_style("top_level_style_adjustments.css")
        + embed_style("parent_logger_style.css")
        + embed_style("child_logger_style.css")
        + embed_style("command_style.css")
        + embed_style("message_style.css")
        + embed_style("detail_list_style.css")
        + embed_style("code_block_style.css")
        + embed_style("output_style.css")
        + embed_style("diagnostics_style.css")
        + embed_style("search_controls.css")
        + embed_script("jquery.slim.min.js")
        + embed_script("bootstrap.bundle.min.js")
        + embed_script("Chart.bundle.min.js")
        + embed_script("search_output.js")
        + embed_html("search_icon.svg")
        + "</head>"
    )


def embed_style(resource: str) -> str:
    """
    Embed a style in the HTML.

    Wrap the given ``resource`` in an appropriate ``<style>...</style>``
    block for embedding in the HTML header.

    Parameters:
        resource:  The name of a style file to embed.

    Returns:
        A string containing the ``<style>...</style>`` block.

    Todo:
      * Should we combine this with :func:`embed_script` and
        :func:`embed_html`?
    """
    return (
        "<style>\n"
        + pkgutil.get_data(__name__, f"resources/{resource}").decode()
        + "\n</style>\n"
    )


def embed_script(resource: str) -> str:
    """
    Embed a script in the HTML.

    Wrap the given ``resource`` in an appropriate
    ``<script>...</script>`` block for embedding in the HTML header.

    Parameters:
        resource:  The name of a script file to embed.

    Returns:
        A string containing the ``<script>...</script>`` block.
    """
    return (
        "<script>\n"
        + pkgutil.get_data(__name__, f"resources/{resource}").decode()
        + "\n</script>\n"
    )


def embed_html(resource: str) -> str:
    """
    Embed the contents of a file in the HTML.

    Get a HTML ``resource`` from a file for the sake of embedding it
    into the HTML header.

    Parameters:
        resource:  The name of a HTML file to embed.

    Returns:
        The contents of the file.

    Todo:
      * Why do we use ``pkgutil.get_data()`` instead of a simple
        ``read()``.
    """
    return pkgutil.get_data(__name__, f"resources/{resource}").decode()


def load_template(template: str) -> str:
    """
    Load a template HTML file.

    Parameters:
        template:  The file name to load.

    Returns:
        A string containing the contents of the file.

    Todo:
      * Should we combine this with :func:`embed_html`?
    """
    template_file = f"resources/templates/{template}"
    return pkgutil.get_data(__name__, template_file).decode()


command_detail_list_template = load_template("command_detail_list.html")
command_detail_template = load_template("command_detail.html")
hidden_command_detail_template = load_template("hidden_command_detail.html")
stat_chart_template = load_template("stat_chart.html")
diagnostics_template = load_template("diagnostics.html")
output_card_template = load_template("output_card.html")
output_card_collapsed_template = load_template("output_card_collapsed.html")
output_block_template = load_template("output_block.html")
output_line_template = load_template("output_line.html")
message_template = load_template("message.html")
html_message_template = load_template("html_message.html")
command_template = load_template("command.html")
child_logger_template = load_template("child_logger.html")
parent_logger_template = load_template("parent_logger.html")



"""Provides the means of collecting various trace data."""

# © 2023 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS).  Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from abc import abstractmethod
from pathlib import Path

from .abstract_method import AbstractMethod


def trace_collector(**kwargs) -> TraceCollector:
    """
    Generate trace collectors.

    A factory method that returns any subclass of
    :class:`TraceCollector` that has the ``@TraceCollector.subclass``
    decorator applied to it.

    Parameters:
        **kwargs:  Any supported arguments of the
            :class:`TraceCollector` subclass.

    Returns:
        A single instance of a :class:`TraceCollector` subclass.
    """
    trace_name = kwargs["trace"]
    collectors = [
        c for c in TraceCollector.subclasses if c.trace_name == trace_name
    ]
    if len(collectors) == 1:
        collector = collectors[0]
        return collector(**kwargs)
    if len(collectors) == 0:
        message = f"Unsupported trace type:  {trace_name}"
        raise RuntimeError(message)
    message = f"Multiple trace types match '{trace_name}'."
    raise RuntimeError(message)


class TraceCollector:
    """
    Trace a command run in the underlying shell.

    Provides an interface for the :class:`ShellLogger` to run commands
    with a certain trace (e.g., ``strace`` or ``ltrace``).
    """

    trace_name = "undefined"  # Should be defined by subclasses.
    subclasses = []  # noqa: RUF012

    @staticmethod
    def subclass(trace_subclass: type):
        """
        Mark a class as being a supported trace collector.

        This is a class decorator that adds to a list of supported
        :class:`TraceCollector` classes for the :func:`trace_collector`
        factory method.
        """
        if issubclass(trace_subclass, TraceCollector):
            TraceCollector.subclasses.append(trace_subclass)
        return trace_subclass

    def __init__(self, **kwargs):
        """
        Initialize the :class:`TraceCollector` object.

        Set up the output file where the trace information will be
        written.
        """
        if kwargs.get("trace_path"):
            self.output_path = Path(kwargs["trace_path"])
        else:
            self.output_path = Path(f"{self.trace_name}.log")

    @property
    @abstractmethod
    def trace_args(self) -> str:
        """
        Get the trace command.

        The trace command and the arguments you pass to it, but not the
        command you're tracing.  E.g., return `strace -f -c -e "open"`.

        Raises:
            AbstractMethod:  This needs to be overridden by subclasses.
        """
        raise AbstractMethod

    def command(self, command: str):
        """
        Get the command to be run in the underlying shell.

        Return a command that runs a trace on ``command``.  E.g., ``ls
        -l`` might get translated to ``strace -f -c -e 'open' -- ls
        -l``.

        Parameters:
            command:  The command to be traced.
        """
        return f"{self.trace_args} -- {command}"


@TraceCollector.subclass
class STraceCollector(TraceCollector):
    """
    Run ``strace`` on commands.

    An interface between :class:`ShellLogger` and the ``strace``
    command.
    """

    trace_name = "strace"

    def __init__(self, **kwargs) -> None:
        """Initialize the :class:`STraceCollector` instance."""
        super().__init__(**kwargs)
        self.summary = bool(kwargs.get("summary"))
        self.expression = kwargs.get("expression")

    @property
    def trace_args(self) -> str:
        """Wraps a command in a ``strace`` command."""
        args = f"strace -f -o {self.output_path}"
        if self.summary:
            args += " -c"
        if self.expression:
            args += f" -e '{self.expression}'"
        return args


@TraceCollector.subclass
class LTraceCollector(TraceCollector):
    """
    Run ``ltrace`` on commands.

    An interface between :class:`ShellLogger` and the ``ltrace``
    command.
    """

    trace_name = "ltrace"

    def __init__(self, **kwargs):
        """Initialize the :class:`LTraceCollector` instance."""
        super().__init__(**kwargs)
        self.summary = bool(kwargs.get("summary"))
        self.expression = kwargs.get("expression")

    @property
    def trace_args(self):
        """Wraps a command in a ``ltrace`` command."""
        args = f"ltrace -C -f -o {self.output_path}"
        if self.summary:
            args += " -c"
        if self.expression:
            args += f" -e '{self.expression}'"
        return args



"""Provides the :class:`AbstractMethod` exception."""

# © 2023 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS).  Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

# SPDX-License-Identifier: BSD-3-Clause

import inspect


class AbstractMethod(NotImplementedError):
    """
    An abstract method meant to be overridden.

    An ``Exception`` denoting an abstract method that is meant to be
    overridden by a subclass.
    """

    def __init__(self):
        """
        Raise a ``NotImplementedError``.

        Indicate which method must be implemented for the class to be
        concrete.
        """
        class_name = (
            inspect.stack()[1].frame.f_locals["self"].__class__.__name__
        )
        method_name = inspect.stack()[1].function
        super().__init__(f"`{class_name}` must implement `{method_name}()`.")



"""Provides the :class:`Shell` class."""

# © 2023 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS).  Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import _thread
import fcntl
import os
import subprocess
import sys
from io import StringIO
from pathlib import Path
from threading import Thread
from time import time
from types import SimpleNamespace
from typing import IO, Optional, TextIO


END_OF_READ = 4


class Shell:
    """
    Manage interactions with the underlying shell.

    Spawns a shell subprocess that inherits five unnamed pipes
    (``stdout``, ``stderr``, ``stdin``, ``aux_stdout``, ``aux_stderr``).

    Attributes:
        login_shell (bool):  Whether or not the spawned shell should be
            a login shell.
        aux_stdin_rfd (int):  The read file descriptor for ``stdin``.
        aux_stdin_wfd (int):  The write file descriptor for ``stdin``.
        aux_stdout_rfd (int):  The read file descriptor for ``stdout``.
        aux_stdout_wfd (int):  The write file descriptor for ``stdout``.
        aux_stderr_rfd (int):  The read file descriptor for ``stderr``.
        aux_stderr_wfd (int):  The write file descriptor for ``stderr``.
        shell_subprocess (Popen[str]):  The subprocess for interacting
            with the shell.
    """

    def __init__(
        self, pwd: Optional[Path] = None, *, login_shell: bool = False
    ) -> None:
        """
        Initialize a :class:`Shell` object.

        Parameters:
            pwd:  The directory to change to when starting the
                :class:`Shell`.
            login_shell:  Whether or not the spawned shell should be a
                login shell.
        """
        self.login_shell = login_shell

        # Corresponds to the 0, 1, and 2 file descriptors of the shell
        # we're going to spawn.
        self.aux_stdin_rfd, self.aux_stdin_wfd = os.pipe()
        self.aux_stdout_rfd, self.aux_stdout_wfd = os.pipe()
        self.aux_stderr_rfd, self.aux_stderr_wfd = os.pipe()

        # Get the current flags of the file descriptors.
        aux_stdout_write_flags = fcntl.fcntl(
            self.aux_stdout_wfd, fcntl.F_GETFL
        )
        aux_stderr_write_flags = fcntl.fcntl(
            self.aux_stderr_wfd, fcntl.F_GETFL
        )

        # Make writes non-blocking.
        fcntl.fcntl(
            self.aux_stdout_wfd,
            fcntl.F_SETFL,
            aux_stdout_write_flags | os.O_NONBLOCK,
        )
        fcntl.fcntl(
            self.aux_stderr_wfd,
            fcntl.F_SETFL,
            aux_stderr_write_flags | os.O_NONBLOCK,
        )

        # Ensure the file descriptors are inheritable by the shell
        # subprocess.
        os.set_inheritable(self.aux_stdout_wfd, True)
        os.set_inheritable(self.aux_stderr_wfd, True)
        shell_command = [os.environ.get("SHELL") or "/bin/sh"]
        if self.login_shell:
            shell_command.append("-l")
        self.shell_subprocess = subprocess.Popen(
            shell_command,
            stdin=self.aux_stdin_rfd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=False,
        )
        os.set_inheritable(self.aux_stdout_wfd, False)
        os.set_inheritable(self.aux_stderr_wfd, False)

        # Start the shell in the given directory.
        if pwd is None:
            pwd = Path.cwd()
        self.cd(pwd)

    def __del__(self) -> None:
        """Close all the open file descriptors."""
        for fd in [
            self.aux_stdin_rfd,
            self.aux_stdin_wfd,
            self.aux_stdout_rfd,
            self.aux_stdout_wfd,
            self.aux_stderr_rfd,
            self.aux_stderr_wfd,
        ]:
            try:
                os.close(fd)
            except OSError as e:
                if "Bad file descriptor" not in e.strerror:
                    raise

    def __eq__(self, other: Shell) -> bool:
        """
        Determine if this :class:`Shell` object is equal to another one.

        Parameters:
            other:  The other :class:`Shell` to compare against.

        Returns:
            ``True`` if the types and working directories of the two
            objects are equal; ``False`` otherwise.
        """
        return isinstance(self, type(other)) and self.pwd() == other.pwd()

    def pwd(self) -> str:
        """
        Get the current working directory.

        Returns:
            The current working directory.
        """
        directory, _ = self.auxiliary_command(posix="pwd", strip=True)
        return directory

    def cd(self, path: Path) -> None:
        """
        Change to the given directory.

        Parameters:
            path:  The directory to change to.
        """
        os.chdir(path)
        self.auxiliary_command(posix=f"cd {path}")

    def run(self, command: str, **kwargs) -> SimpleNamespace:
        """
        Run a command in the underlying shell.

        Write a ``command`` to the :class:`Shell` class' shell
        subprocess' ``stdin``, and pull the ``stdout`` and ``stderr``.

        Parameters:
            command:  The command to run in the shell subprocess.
            **kwargs:  Any additional arguments to pass to :func:`tee`.

        Returns:
            The command run, along with its return code, ``stdout``,
            ``stderr``, start/stop time, and duration.
        """
        milliseconds_per_second = 10**3
        start = round(time() * milliseconds_per_second)

        # Wrap the `command` in {braces} to support newlines and
        # heredocs to tell the shell "this is one giant statement".
        # Then run the command.
        if kwargs.get("devnull_stdin"):
            os.write(
                self.aux_stdin_wfd, f"{{\n{command}\n}} </dev/null\n".encode()
            )
        else:
            os.write(self.aux_stdin_wfd, f"{{\n{command}\n}}\n".encode())

        # Set the `RET_CODE` environment variable, such that we can
        # access it later.
        os.write(self.aux_stdin_wfd, "RET_CODE=$?\n".encode())

        # Because these writes are non-blocking, tell the shell that the
        # writes are complete.
        os.write(self.aux_stdin_wfd, "printf '\\4'\n".encode())
        os.write(self.aux_stdin_wfd, "printf '\\4' 1>&2\n".encode())

        # Tee the output to multiple sinks (files, strings,
        # `stdout`/`stderr`).
        try:
            output = self.tee(
                self.shell_subprocess.stdout,
                self.shell_subprocess.stderr,
                **kwargs,
            )

        # Note:  If something goes wrong in `tee()`, the only way to reliably
        # propagate an exception from a thread that's spawned is to raise a
        # KeyboardInterrupt.
        except KeyboardInterrupt:
            os.close(self.aux_stdin_wfd)
            message = (
                f"There was a problem running the command `{command}`.  "
                "This is a fatal error and we cannot continue.  Ensure that "
                "the syntax of the command is correct."
            )
            raise RuntimeError(message) from None
        finish = round(time() * milliseconds_per_second)

        # Pull the return code and return the results.  Note that if the
        # command executed spawns a sub-shell, you won't really have a
        # return code.
        aux_out, _ = self.auxiliary_command(posix="echo $RET_CODE")
        try:
            return_code = int(aux_out)
        except ValueError:
            return_code = "N/A"
        return SimpleNamespace(
            returncode=return_code,
            args=command,
            stdout=output.stdout_str,
            stderr=output.stderr_str,
            start=start,
            finish=finish,
            wall=finish - start,
        )

    @staticmethod
    def tee(  # noqa: C901
        stdout: Optional[IO[bytes]], stderr: Optional[IO[bytes]], **kwargs
    ) -> SimpleNamespace:
        """
        Write output/error streams to multiple files.

        Split ``stdout`` and ``stderr`` file objects to write to
        multiple files.

        Parameters:
            stdout:  The ``stdout`` file object to be split.
            stderr:  The ``stderr`` file object to be split.
            **kwargs:  Additional arguments.

        Returns:
            The ``stdout`` and ``stderr`` as strings.

        Todo:
          * Replace ``**kwargs`` with function arguments.
        """
        sys_stdout = None if kwargs.get("quiet_stdout") else sys.stdout
        sys_stderr = None if kwargs.get("quiet_stderr") else sys.stderr
        stdout_io = StringIO() if kwargs.get("stdout_str") else None
        stderr_io = StringIO() if kwargs.get("stderr_str") else None
        stdout_path = kwargs.get("stdout_path", Path(os.devnull)).open("a")
        stderr_path = kwargs.get("stderr_path", Path(os.devnull)).open("a")
        stdout_tee = [sys_stdout, stdout_io, stdout_path]
        stderr_tee = [sys_stderr, stderr_io, stderr_path]

        def write(input_file: TextIO, output_files: list[TextIO]) -> None:
            """
            Write an input to multiple outputs.

            Take the data from an input file object and write it to
            multiple output file objects.

            Parameters:
                input_file:  The file object from which to read.
                output_files:  A list of file objects to write to.
            """
            # Read chunks from the input file.
            chunk_size = 4096  # 4 KB
            chunk = os.read(input_file.fileno(), chunk_size)
            while chunk and chunk[-1] != END_OF_READ:
                for output_file in output_files:
                    if output_file is not None:
                        output_file.write(chunk.decode(errors="ignore"))
                chunk = os.read(input_file.fileno(), chunk_size)

            # If something goes wrong in the `tee()`, the only way to
            # reliably propagate an exception from a thread that's
            # spawned is to raise a KeyboardInterrupt.
            if not chunk:
                _thread.interrupt_main()

            # Remove the end-of-transmission character, and write the
            # last chunk.
            chunk = chunk[:-1]
            for output_file in output_files:
                if output_file is not None:
                    output_file.write(chunk.decode(errors="ignore"))

        # Spawn threads to write to `stdout` and `stderr`.
        threads = [
            Thread(target=write, args=(stdout, stdout_tee)),
            Thread(target=write, args=(stderr, stderr_tee)),
        ]
        for thread in threads:
            thread.daemon = True
            thread.start()
        for thread in threads:
            thread.join()
        stdout_str = stdout_io.getvalue() if stdout_io is not None else None
        stderr_str = stderr_io.getvalue() if stderr_io is not None else None

        # Close any open file descriptors and return the `stdout` and
        # `stderr`.
        for file in stdout_tee + stderr_tee:
            if (
                file not in [None, sys.stdout, sys.stderr, sys.stdin]
                and not file.closed
            ):
                file.close()
        return SimpleNamespace(stdout_str=stdout_str, stderr_str=stderr_str)

    def auxiliary_command(
        self, **kwargs
    ) -> tuple[Optional[str], Optional[str]]:
        """
        Run auxiliary commands like `umask`, `pwd`, `env`, etc.

        Parameters:
            **kwargs:  Additional arguments.

        Note:  This is effectively the same as :func:`run`, but:
            1. The ``stdout`` and ``stderr`` get redirected to the
               auxiliary file descriptors.
            2. You don't tee the ``stdout`` or ``stderr``.

        Todo:
          * Maybe combine this with :func:`run` with extra flags.
          * Replace ``**kwargs`` with function arguments.

        Returns:
            The ``stdout`` and ``stderr`` of the command run.
        """
        stdout, stderr = None, None
        out = self.aux_stdout_wfd
        err = self.aux_stderr_wfd
        if os.name in kwargs:
            cmd = kwargs[os.name]
            os.write(self.aux_stdin_wfd, f"{cmd} 1>&{out} 2>&{err}\n".encode())
            os.write(self.aux_stdin_wfd, f"printf '\\4' 1>&{out}\n".encode())
            os.write(self.aux_stdin_wfd, f"printf '\\4' 1>&{err}\n".encode())
            stdout = ""
            stderr = ""

            max_anonymous_pipe_buffer_size = 65536
            aux = os.read(self.aux_stdout_rfd, max_anonymous_pipe_buffer_size)
            while aux[-1] != END_OF_READ:
                stdout += aux.decode()
                aux = os.read(
                    self.aux_stdout_rfd, max_anonymous_pipe_buffer_size
                )
            aux = aux[:-1]
            stdout += aux.decode()
            aux = os.read(self.aux_stderr_rfd, max_anonymous_pipe_buffer_size)
            while aux[-1] != END_OF_READ:
                stderr += aux.decode()
                aux = os.read(
                    self.aux_stderr_rfd, max_anonymous_pipe_buffer_size
                )
            aux = aux[:-1]
            stderr += aux.decode()
            if kwargs.get("strip"):
                if stdout:
                    stdout = stdout.strip()
                if stderr:
                    stderr = stderr.strip()
        return stdout, stderr



"""Provides the various means of collecting machine statistics."""

# © 2023 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS).  Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import os
from abc import abstractmethod
from multiprocessing import Manager, Process
from pathlib import Path
from time import sleep, time
from typing import TYPE_CHECKING

from .abstract_method import AbstractMethod

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager

try:
    import psutil
except ModuleNotFoundError:
    psutil = None


def stats_collectors(**kwargs) -> list[StatsCollector]:
    """
    Generate stats collectors.

    A factory method that returns a list of any subclasses of
    :class:`StatsCollector` that have the ``@StatsCollector.subclass``
    decorator applied to them.

    Parameters:
        **kwargs:  Any supported arguments of the
            :class:`StatsCollector` subclasses.

    Returns:
        A collection of instances of :class:`StatsCollector` subclasses.
    """
    collectors = []
    if "measure" in kwargs:
        interval = kwargs.get("interval", 1.0)
        manager = Manager()
        for collector in StatsCollector.subclasses:
            if collector.stat_name in kwargs["measure"]:
                collectors.append(collector(interval, manager))
    return collectors


class StatsCollector:
    """
    Collect statistics while running command in the shell.

    Provides an interface for the :class:`ShellLogger` to run commands
    while collecting various system statistics.
    """

    stat_name = "undefined"  # Should be defined by subclasses.
    subclasses = []  # noqa: RUF012

    @staticmethod
    def subclass(stats_collector_subclass: type):
        """
        Mark a class as being a supported stats collector.

        This is a class decorator that adds to a list of supported
        :class:`StatsCollector` classes for the :func:`stats_collectors`
        factory method.
        """
        if issubclass(stats_collector_subclass, StatsCollector):
            StatsCollector.subclasses.append(stats_collector_subclass)
        return stats_collector_subclass

    def __init__(self, interval: float, _: SyncManager):
        """
        Initialize the :class:`StatsCollector` object.

        Set the poling interval, and create the process for collecting
        the statistics.

        Parameters:
            interval:  How long to sleep between collecting statistics.

        Note:
            A ``SyncManager`` will be supplied at the second argument
            for any subclasses.
        """
        self.interval = interval
        self.process = Process(target=self.loop, args=())

    def start(self):
        """
        Start a subprocess.

        Poll at a certain interval for certain statistics.
        """
        self.process.start()

    def loop(self):
        """
        Loop while collecting statistics.

        Infinitely loop, collecting statistics, until the subprocess is
        terminated.
        """
        while True:
            self.collect()
            sleep(self.interval)

    @abstractmethod
    def collect(self):
        """
        Instantaneously collect a statistic.

        This is meant to be called repeatedly after some time interval.

        Raises:
            AbstractMethod:  This must be overridden by subclasses.
        """
        raise AbstractMethod

    @abstractmethod
    def unproxied_stats(self):
        """
        Convert to standard Python data types.

        Convert from Python's Manager's data structures to base Python
        data structures.

        Raises:
            AbstractMethod:  This must be overridden by subclasses.
        """
        raise AbstractMethod

    def finish(self):
        """
        Stop collecting statistics.

        Terminate the infinite loop that's collecting the statistics,
        and then return the unproxied statistics.
        """
        self.process.terminate()
        return self.unproxied_stats()


if psutil is not None:

    @StatsCollector.subclass
    class DiskStatsCollector(StatsCollector):
        """
        Collect disk usage statistics.

        A means of running commands while collecting disk usage
        statistics.
        """

        stat_name = "disk"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the :class:`DiskStatsCollector` object.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)
            self.stats = manager.dict()
            self.mount_points = [
                p.mountpoint for p in psutil.disk_partitions()
            ]
            for location in [
                "/tmp",
                "/dev/shm",
                f"/var/run/user/{os.getuid()}",
            ]:
                if (
                    location not in self.mount_points
                    and Path(location).exists()
                ):
                    self.mount_points.append(location)
            for m in self.mount_points:
                self.stats[m] = manager.list()

        def collect(self) -> None:
            """Poll the disks to determine how much free space they have."""
            milliseconds_per_second = 10**3
            timestamp = round(time() * milliseconds_per_second)
            for m in self.mount_points:
                self.stats[m].append((timestamp, psutil.disk_usage(m).percent))

        def unproxied_stats(self) -> dict:
            """
            Convert the statistics to standard Python data types.

            Translate the statistics from the multiprocessing
            ``SyncManager`` 's data structure to a ``dict``.

            Returns:
                A mapping from the disk mount points to tuples of
                timestamps and percent of disk space free.
            """
            return {k: list(v) for k, v in self.stats.items()}

    @StatsCollector.subclass
    class CPUStatsCollector(StatsCollector):
        """
        Collect CPU statistics.

        A means of running commands while collecting CPU usage
        statistics.
        """

        stat_name = "cpu"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the :class:`CPUStatsCollector` object.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)
            self.stats = manager.list()

        def collect(self) -> None:
            """Determine how heavily utilized the CPU is at the moment."""
            milliseconds_per_second = 10**3
            timestamp = round(time() * milliseconds_per_second)
            self.stats.append((timestamp, psutil.cpu_percent(interval=None)))

        def unproxied_stats(self) -> list[tuple[float, float]]:
            """
            Convert the statistics to standard Python data types.

            Translate the statistics from the multiprocessing
            ``SyncManager`` 's data structure to a ``list``.

            Returns:
                A list of (timestamp, % CPU used) data points.
            """
            return list(self.stats)

    @StatsCollector.subclass
    class MemoryStatsCollector(StatsCollector):
        """
        Collect memory statistics.

        A means of running commands while collecting memory usage
        statistics.
        """

        stat_name = "memory"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the :class:`MemoryStatsCollector` object.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)
            self.stats = manager.list()

        def collect(self) -> None:
            """Determine how much memory is currently being used."""
            milliseconds_per_second = 10**3
            timestamp = round(time() * milliseconds_per_second)
            self.stats.append((timestamp, psutil.virtual_memory().percent))

        def unproxied_stats(self) -> list[tuple[float, float]]:
            """
            Convert the statistics to standard Python data types.

            Translate the statistics from the multiprocessing
            ``SyncManager`` 's data structure to a ``list``.

            Returns:
                A list of (timestamp, % memory used) data points.
            """
            return list(self.stats)


# If we don't have `psutil`, return null objects.
else:

    @StatsCollector.subclass
    class DiskStatsCollector(StatsCollector):
        """
        A null disk statistics collector for when data aren't available.

        A phony :class:`DiskStatsCollector` used when ``psutil`` is
        unavailable.  This collects no disk statistics.
        """

        stat_name = "disk"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the object via the parent's constructor.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)

        def collect(self) -> None:
            """Don't collect any disk statistics."""
            pass

        def unproxied_stats(self) -> None:
            """
            If asked for the disk statistics, don't provide any.

            Returns:
                None
            """
            return

    @StatsCollector.subclass
    class CPUStatsCollector(StatsCollector):
        """
        A null CPU statistics collector for when data aren't available.

        A phony :class:`CPUStatsCollector` used when ``psutil`` is
        unavailable.  This collects no CPU statistics.
        """

        stat_name = "cpu"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the object via the parent's constructor.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)

        def collect(self) -> None:
            """Don't collect any CPU statistics."""
            pass

        def unproxied_stats(self) -> None:
            """
            If asked for CPU statistics, don't provide any.

            Returns:
                None
            """
            return

    @StatsCollector.subclass
    class MemoryStatsCollector(StatsCollector):
        """
        A null memory stats collector for when data aren't available.

        A phony :class:`MemoryStatsCollector` used when ``psutil`` is
        unavailable.  This collects no memory statistics.
        """

        stat_name = "memory"

        def __init__(self, interval: float, manager: SyncManager) -> None:
            """
            Initialize the object via the parent's constructor.

            Parameters:
                interval:  How many seconds to sleep between polling.
                manager:  The multiprocessing manager used to control
                    the process used to collect the statistics.
            """
            super().__init__(interval, manager)

        def collect(self) -> None:
            """Don't collect any memory statistics."""
            pass

        def unproxied_stats(self) -> None:
            """
            If asked for memory statistics, don't provide any.

            Returns:
                None
            """
            return
